## Copyright (c) 2014 David Owen.  All rights reserved.

library(rgl)
library(rpart)
library(MASS)
par(bg="black", col.axis="white", col.lab="white", col.main="white", col.sub="white", fg="gray")
plot(0,0)
title("Classification")

clamp.0.1 <- function (x) {
    pmax(pmin(x, 1), 0)
}

class.ramp <- function (x, alpha=1) {
    if(is.factor(x)) x <- as.integer(x) - 1
    pri <- clamp.0.1(abs(x)/2 - abs(x-1)/2 + 0.5)
    sec <- clamp.0.1(-1*abs(x) + 1*abs(x-1) + x)
    rgb(sec, pri, sec, alpha)
}

class3.ramp <- function (x, alpha=1) {
    if(is.factor(x)) x <- as.integer(x) - 1
    ifelse(x==0, rgb(1, 0, 1, alpha), ifelse(x==1, rgb(0, 1, 0, alpha), ifelse(x==2, rgb(0.8, 0.8, 0, alpha), rgb(0.7, 0.7, 0.7, alpha))))
}

psline <- function (x, dx, ...) {
    d <- x[1]/dx[1]
    int <- x[2] - d*dx[2]
    abline(int, dx[2]/dx[1], ...)
}

rotate <- function (x, theta) {
    y <- c(-x[2], x[1])
    cos(theta)*x + sin(theta)*y
}

fancyhist <- function (points, classes, boundary=T) {
    h <- hist(points, breaks=30, plot=F)
    hist(points, breaks=h$breaks, xlab="Projection", main=NA, border="black")
    hist(points[classes==1], breaks=h$breaks, freq=T, add=T,
         border=class.ramp(1), col=class.ramp(1, 0.5), density=20, angle=-45)
    hist(points[classes==0], breaks=h$breaks, freq=T, add=T,
         border=class.ramp(0), col=class.ramp(0, 0.5), density=20)
    if(boundary) points(0.5, 0, col="blue", pch=19)
}

## Welcome to classification.  We're going to discuss the geometry and
## evolution of classification, starting with a very basic linear
## classifier.

## First, we'll make some data.  It's going to be divided into 2
## classes.
d <- data.frame(x=runif(100), y=0)
d$c <- round(runif(100))
d$y <- 0.5*d$x - 1 + 0.5*d$c + runif(100, -0.1, 0.1)
plot(d$x, d$y, col=class.ramp(d$c))

## These points are class 0, these are class 1.  Our goal is to
## construct a very simple linear classifier that, when given a point
## anywhere on this grid, will correctly classify it into either group
## 0 or group 1.

## To do that, we'll just use a linear model in R.  This is what it
## is: c ~ x + y, which means to predict c from x and y.  We're not
## adding x and y, just including each in the model.

m <- lm(c ~ x + y, d)
summary(m)

## The summary of the model shows an excellent fit.  Each coefficient
## has very high t-values and very low p-values, and an overall
## goodness of fit as signified by the F statistic.

## To see what this classifier function looks like, we'll throw out a
## bunch of random points on this grid and run them through the
## classifier.

t <- data.frame(x=runif(1000), y=runif(1000, -1, 0.5))
t$c <- predict(m, newdata=t)
plot(t$x, t$y, col=class.ramp(t$c, 3/4))
points(d$x, d$y, col=class.ramp(d$c), pch=19)

## Here's what we get.  We see that the gradient is near 0 in the
## middle of the 0 class, and near 1 in the middle of the 1 class.

## So, the linear classifier describes a gradient on which we can rank
## points into each class according to their level on that gradient.

## This is an interesting geometric observation.  This classifier
## basically works with contours on the gradient, like a topographic
## map.  If we were to look at it from the side, it might look like
## this.  (slope with points near the slope).

bg3d(color="black")
view3d(0,0, fov=0)
plot3d(t$x, t$y, t$c, col=class.ramp(t$c), size=1)
points3d(d$x, d$y, d$c, col=class.ramp(d$c))
axes3d(color="white")

## The 0-contour runs here, the 1-contour runs here.  Because I've
## create such a nice data set, it happens that we can use the 0.5
## contour as being the decision boundary.  Anything less is called a
## 0, anything greater is called a 1.

abclines3d(0, (0.5-coef(m)[1]) / coef(m)[3], 0.5, -coef(m)[3], coef(m)[2], 0, lw=5, col="blue")

## Another way to look at the geometry.

cx <- c(mean(d$x), mean(d$y))
plot(d$x, d$y, col=class.ramp(d$c), asp=1)
abline((0.5-coef(m)[1]) / coef(m)[3], -coef(m)[2]/coef(m)[3], col="blue", lw=2)

## Focus on this line (0.5 contour) as separating our space.  Anything
## on this side is one class, on the other the other.  Imagine we had
## more dimensions to our space.  The separating boundary isn't just a
## line, it would be a 2-dim plane in 3-dim space, or generally an n-1
## plane in n-dim space, so that it divides the space in two.

## Here's the cool math stuff.  A plane and a line are dual concepts.
## They complement each other and are almost the same thing.  A plane
## can be defined in terms of a point on that plane and a vector that
## is perpendicular to that plane.

## A point and a vector also define a line, a line perpendicular to
## the plane.  So, the dual of an n-1 dim plane is a 1-dim line.

## So, this boundary is either defined as an n-1 dim plane, or
## equivalently a line perpendicular to it.  So, in terms of the
## plane, we can plug any point into the equation of the plane.  If
## the equation yields 0, it's on the plane; negative on one side,
## positive on the other.

psline(cx, c(coef(m)[2], coef(m)[3]))

## Equivalently, we can plug a point into the equation of the line to
## project that point perpendicularly onto the line.  The equation
## yields how far, in terms of the defining vector, our point is from
## the line's reference point.

## Being able to separate or classify points using either a plane or a
## line is linear separability.

## Let's take the case where we project onto the line.  As we project
## onto the line, we get these sets of points.  (It would be awesome
## to have points added incrementally to both the overhead and the
## side views.)  (Or just rotate interactively in gnuplot.)

plot(predict(m), 0*predict(m), pch=" ", xlab="Projection", ylab="")
abline(0, 0, col="white")
points(predict(m), 0*predict(m), col=class.ramp(d$c))
points(0.5, 0, col="blue", pch=19)

## Let's create a vertical axis on this display and graph the
## histograms of the classes there.

fancyhist(predict(m), d$c)

## So, with the simple linear classifier, we've found the view that
## separates these two distributions the most.  Then we see that
## anything less than 0.5 is probably in this distribution, anything
## greater probably in this.

## Let's consider what the linear regression we're using is actually
## doing, by looking at some different projections.

v <- c(coef(m)[2], coef(m)[3])
plot(d$x, d$y, col=class.ramp(d$c), asp=1)
psline(cx, v)

# Rotate it 15 degrees

v2 <- rotate(v, pi/12)
plot(d$x, d$y, col=class.ramp(d$c), asp=1)
psline(cx, v2)

p <- v2[1]*d$x + v2[2]*d$y
p <- p - mean(p[d$c==0])
p <- p / mean(p[d$c==1])
fancyhist(p, d$c)

# Now 30 degrees

v2 <- rotate(v, 2*pi/12)
plot(d$x, d$y, col=class.ramp(d$c), asp=1)
psline(cx, v2, col="white")

p <- v2[1]*d$x + v2[2]*d$y
p <- p - mean(p[d$c==0])
p <- p / mean(p[d$c==1])
fancyhist(p, d$c)

# Now 45.

v2 <- rotate(v, 3*pi/12)
plot(d$x, d$y, col=class.ramp(d$c), asp=1)
psline(cx, v2)

p <- v2[1]*d$x + v2[2]*d$y
p <- p - mean(p[d$c==0])
p <- p / mean(p[d$c==1])
fancyhist(p, d$c)

## So, the linear regression is actually finding the best rotation, or
## angle, along which to view the data to get the best separation.  It
## does this by minimizing the sum variance of each class around its
## respective level (0 or 1).

## Let's look at this in terms of transformations.

system("evince -s transform.pdf")

## The linear equation is an arbitrary transform followed by a
## projection onto a single axis!  In other words, linear regression
## will translate, scale, even rotate the data to find the view that
## gives the best separation, determined by minimizing the total
## variance.  A linear regression is the same as a projected
## transform.

## Back to distributions.  Here's the distribution we already saw.

fancyhist(predict(m), d$c)

## Looking at this "from the side" again, we'll see that we've been
## fitting a line through the two classes.

plot(predict(m), d$c, col=class.ramp(d$c), xlab="X", ylab="Class")
abline(0, 1)
points(0.5, 0.5, col="blue", pch=19)

## Now, we don't always have nice separate distributions like these.
## We've been able to get away with this because the two distributions
## are narrow enough to not spill over the 0.5 boundary.  That's
## seldom the case, though, and we need a more sophisticated way to
## determine that boundary.

# Here's an example from the "Iris" data.
data(iris)
d <- iris[iris$Species %in% c("versicolor", "virginica"),]
d$Species <- as.factor(as.character(d$Species))
d <- d[order(d$Petal.Length),]
d$c <- as.integer(d$Species) - 1
plot(d$Petal.Length, d$c, col=class.ramp(d$c, 0.8))

## How do we know where the decision boundary is?  There are two basic
## approaches.  The first is to fit a "probit" function.

t <- data.frame(Petal.Length=seq(3, 7, length.out=200))
m <- glm(c ~ Petal.Length, d, family=binomial("probit"))
lines(t$Petal.Length, predict(m, newdata=t, type="r"))

## If your toolkit doesn't include probit, a "logit" is a very close
## approximation.  You can implement it by tweaking your class
## variable slightly away from 0 and 1, applying the logistic
## transform, performing a linear regression, then applying the
## reverse transform.  The result looks like this:

m <- glm(c ~ Petal.Length, d, family=binomial("logit"))
lines(t$Petal.Length, predict(m, newdata=t, type="r"), lty="FF")

## These regressions represent the probability that a point is in one
## class, so 0.5, being 50%, is always the boundary.

## The second approach to finding the boundary is to use decision
## trees.  Building one on the iris data gives this tree:

m <- rpart(c ~ Petal.Length, d, method="class")
plot(m, margin=0.2)
text(m, use.n=T, fancy=T)

## And that gives this classification result:

plot(d$Petal.Length, jitter(predict(m)[,2]), col=class.ramp(d$c), pch=21, bg=class.ramp(d$c, 0.33))

## With a little more detail:

m <- rpart(c ~ Petal.Length, d)
plot(d$Petal.Length, jitter(predict(m)), col=class.ramp(d$c), pch=21, bg=class.ramp(d$c, 0.33))

m <- rpart(c ~ Petal.Length, d, control=rpart.control(cp=0.001))
plot(d$Petal.Length, jitter(predict(m)), col=class.ramp(d$c), pch=21, bg=class.ramp(d$c, 0.33))

## You can actually approximate the probit using decision trees, if
## the data follows that model.  Decision trees can handle arbitrary
## distributions, though.  For example, probit and logit couldn't
## handle this very well, but a decision tree could:

x1 <- rnorm(1000, 0, 0.8)
x2 <- rnorm(300, 1, 0.2)
fancyhist(c(x1,x2), c(0*x1, 1+0*x2), boundary=F)

## A decision-tree handles this just fine:

d <- data.frame(x=c(x1, x2), c=c(0*x1, 1+0*x2))
m <- rpart(c ~ x, d)
plot(d$x, jitter(predict(m), 10), col=class.ramp(d$c, 0.33))

## Going back to linear separability.  Sometimes we have data that
## isn't linearly separable, like this:

d <- data.frame(x=runif(1000), y=runif(1000))
d$c <- sqrt((d$x-0.7)^2 + (d$y-0.4)^2) < 0.2
plot(d$x, d$y, col=class.ramp(d$c), asp=T)

## Try the simple model:
m <- glm(c ~ x + y, d, family=binomial("probit"))
t <- data.frame(x=runif(1000,0,2), y=runif(1000,-1,1))
plot(t$x, t$y, col=class.ramp(predict(m, newdata=t), 0.3), pch=19, asp=T)
points(d$x, d$y, col=class.ramp(d$c))
t <- data.frame(x=runif(1000), y=runif(1000))

## Let's look at an interesting "transform":
d$xx <- d$x^2
d$yy <- d$y^2
plot(d$xx - 1.4*d$x + 0.49, d$yy - 0.8*d$y + 0.16, col=class.ramp(d$c), asp=T)

## Zoomed in:
plot(d$xx - 1.4*d$x + 0.49, d$yy - 0.8*d$y + 0.16, col=class.ramp(d$c), xlim=c(0,0.08), ylim=c(0,0.08), asp=T)

## Linear separability!  Here are the predictions:
m <- glm(c ~ x + y + xx + yy, d, family=binomial("probit"))
t$xx <- t$x^2
t$yy <- t$y^2
plot(t$x, t$y, col=class.ramp(predict(m, newdata=t, type="r"), 0.9), pch=1, asp=T)

## And now I'll add the actuals back in.
points(d$x, d$y, col=class.ramp(d$c, 0.3), pch=19)

## To handle a decision boundary in the shape of any conic
## cross-section, add in all 2nd-order terms.

0

## Part 2

## K-class classification (2 dimensional).  Can't use a single linear
## classifier.  Things might not line up.

## One approach is to build several linear classifiers and combine
## them together in some useful way.

## First, 1-v-a.

d <- data.frame(iris[,c(2, 4, 5)])
d$Sepal.Width <- jitter(d$Sepal.Width, 2)
d$Petal.Width <- jitter(d$Petal.Width, 2)
d$c1 <- d$Species == "setosa"
d$c2 <- d$Species == "versicolor"
d$c3 <- d$Species == "virginica"
d

plot(d[,1:2], col=class3.ramp(d$Species))

m1 <- glm(c1 ~ Sepal.Width + Petal.Width, binomial("probit"), d)
m2 <- glm(c2 ~ Sepal.Width + Petal.Width, binomial("probit"), d)
m3 <- glm(c3 ~ Sepal.Width + Petal.Width, binomial("probit"), d)

plot(d[,1:2], col=class.ramp(predict(m1, type="r")))

plot(d[,1:2], col=class.ramp(predict(m2, type="r")))

plot(d[,1:2], col=class.ramp(predict(m3, type="r")))

t <- data.frame(Sepal.Width=runif(1000, 2, 4.5), Petal.Width=runif(1000, 0, 2.5))

## To combine the classifiers, we can "cast votes":

p1 <- 3*round(predict(m1, t, type="r")) - 1
p2 <- 3*round(predict(m2, t, type="r")) - 1
p3 <- 3*round(predict(m3, t, type="r")) - 1
pp1 <- p1 > p2 & p1 > p3
pp2 <- p2 > p1 & p2 > p3
pp3 <- p3 > p1 & p3 > p2
t$c <- ifelse(pp1, 0, ifelse(pp2, 1, ifelse(pp3, 2, 3)))
plot(t[,1:2], col=class3.ramp(t$c, 0.67))
points(d[,1:2], col=class3.ramp(d$Species), pch=19)

## Better than simple majority is adding up the values from the
## probit/logit directly.

p1 <- 3*predict(m1, t, type="r") - 1
p2 <- 3*predict(m2, t, type="r") - 1
p3 <- 3*predict(m3, t, type="r") - 1
pp1 <- p1 > p2 & p1 > p3
pp2 <- p2 > p1 & p2 > p3
pp3 <- p3 > p1 & p3 > p2
t$c <- ifelse(pp1, 0, ifelse(pp2, 1, ifelse(pp3, 2, 3)))
plot(t[,1:2], col=class3.ramp(t$c, 0.67))
points(d[,1:2], col=class3.ramp(d$Species), pch=19)

## Or even, a decision tree.  In this case, a linear classifer (as
## opposed to probit/logit) is better.

m1 <- lm(c1 ~ Sepal.Width + Petal.Width, d)
m2 <- lm(c2 ~ Sepal.Width + Petal.Width, d)
m3 <- lm(c3 ~ Sepal.Width + Petal.Width, d)
d$p1 <- predict(m1)
d$p2 <- predict(m2)
d$p3 <- predict(m3)
m <- rpart(Species ~ p1 + p2 + p3, d)
t$c <- predict(m, t, type="c")
plot(t[,1:2], col=class3.ramp(t$c, 0.67))
points(d[,1:2], col=class3.ramp(d$Species), pch=19)

## The downside of 1-v-a is that if you get a class situated between
## other classes, it might have almost no distinguishing power at all.

## Now for 1-v-1.

m12 <- glm(c1 ~ Sepal.Width + Petal.Width, binomial("probit"), d[d$c1 | d$c2,])
m13 <- glm(c1 ~ Sepal.Width + Petal.Width, binomial("probit"), d[d$c1 | d$c3,])
m23 <- glm(c2 ~ Sepal.Width + Petal.Width, binomial("probit"), d[d$c2 | d$c3,])

plot(d[,1:2], col=class.ramp(predict(m12, d, type="r")))

plot(d[,1:2], col=class.ramp(predict(m13, d, type="r")))

plot(d[,1:2], col=class.ramp(predict(m23, d, type="r")))

p12 <- round(predict(m12, t, type="r"))
p13 <- round(predict(m13, t, type="r"))
p23 <- round(predict(m23, t, type="r"))
pp1 <- p12 & p13
pp2 <- !p12 & p23
pp3 <- !p13 & !p23
t$c <- ifelse(pp1, 0, ifelse(pp2, 1, ifelse(pp3, 2, 3)))
plot(t[,1:2], col=class3.ramp(t$c, 0.67))
points(d[,1:2], col=class3.ramp(d$Species), pch=19)

## 1-v-all and 1-v-1 are just approximations of the ideal situation.
## If we go back to the idea of projecting distributions, we can
## attack the problem of projecting them to get the most separation.
## Fisher tackled this problem, and called it Discriminant Analysis.

system("evince -s lda.pdf")

#system("evince -s white.pdf")

## LDA gives us k-1 projections for k classes.

## (Talk about arranging k classes on independent axes: k-1 axes is
## sufficient, and sometimes we can do with even less.)

d <- iris[iris$Species %in% c("versicolor", "virginica"), c(2, 4, 5)]
d[,1] <- jitter(d[,1])
d[,2] <- jitter(d[,2])
d[,3] <- as.factor(as.character(d[,3]))
plot(d[,1:2], col=class.ramp(d$Species), asp=T)

m <- lda(Species ~ ., d)
plot(predict(m)$x, runif(100, -0.1, 0.1), col=class.ramp(d$Species), xlab="LD1", ylab="Jitter", ylim=c(-1, 1))

fancyhist(predict(m)$x, 1*(d$Species == "virginica"), F)

## This contrasts with PCA or SVD, which finds the axes through which
## the data as a whole has the most variation, with all axes being
## orthogonal to each other.  Here's an example where PCA would be no
## help in classification:

m <- prcomp(d[,1:2])
t <- data.frame(predict(m))
t$c <- 1*(d$Species == "virginica")
plot(t[,1:2], col=class.ramp(t$c), asp=T)

fancyhist(t[,1], t$c, F)

fancyhist(t[,2], t$c, F)

## So, here's LDA on the full iris data:

d <- data.frame(iris[,c(2,4,5)])
d$Sepal.Width <- jitter(d$Sepal.Width, 2)
d$Petal.Width <- jitter(d$Petal.Width, 2)
m <- lda(Species ~ ., d)
plot(predict(m)$x, col=class3.ramp(d$Species))

## And here are what the decision boundaries look like in
## untransformed space:

sel <- function(x) { runif(1000, min(x), max(x)) }

t <- data.frame(Sepal.Length=sel(iris$Sepal.Length), Sepal.Width=sel(iris$Sepal.Width), Petal.Length=sel(iris$Petal.Length), Petal.Width=sel(iris$Petal.Width))
t$Species <- predict(m, t)$class
plot(t[,c(2,4)], col=class3.ramp(t$Species, 0.67))
points(d[,1:2], col=class3.ramp(d$Species), pch=19)

## If the separating boundaries are not linear for your data, 2
## options: QDA, which lifts some of the Gaussian parameter
## assumptions, or adding some or all the 2-order terms into your
## model.  Validate!

m <- qda(Species ~ ., d)
t$Species <- predict(m, t)$class
plot(t[,c(2,4)], col=class3.ramp(t$Species, 0.67))
points(d[,1:2], col=class3.ramp(d$Species), pch=19)

## Yay!  You graduated!  All the methods we discussed are basically
## projections onto lines, so we can now take high-dimensional spaces
## and reduce them down in some useful way.  So, let's look at some
## high-dimensional stuff.

load("~/2/samsungData.rda")
samsungData <- samsungData[,c(1:561,563)]
names(samsungData) <- paste(1:562, names(samsungData), sep=".")
names(samsungData)[562] <- "activity"
samsungData$activity <- as.factor(samsungData$activity)

## Samsung device sensor data collected from 30 people for 6
## activities: laying, sitting, standing, walking on level ground,
## walking up stairs, walking down stairs.  7300 observations of 560
## variables.

r <- runif(5293)
d <- samsungData[r < 0.5,]
dd <- samsungData[r >= 0.5,]

## For some apples-to-apples comparisons, we'll use rpart on the first
## few examples.  First, axis-aligned decision trees:

m.aa <- rpart(activity ~ ., d, method="class")
sum(predict(m.aa, dd, type="class") != dd$activity)

## PCA:

pca <- prcomp(d[,1:561])
t <- data.frame(predict(pca))
t$activity <- d$activity
m.pca <- rpart(activity ~ ., t, method="class")
sum(predict(m.pca, data.frame(predict(pca, dd)), type="class") != dd$activity)

t <- data.frame(predict(pca)[,1:5])
t$activity <- d$activity
m.pca <- rpart(activity ~ ., t, method="class")
sum(predict(m.pca, data.frame(predict(pca, dd)), type="class") != dd$activity)

## LDA w/decision-tree

ld <- lda(activity ~ ., d)
t <- data.frame(predict(ld)$x)
t$activity <- d$activity
m.lda <- rpart(activity ~ ., t)
sum(predict(m.lda, data.frame(predict(ld, dd)$x), type="class") != dd$activity)

## LDA itself
sum(predict(ld, dd)$class != dd$activity)

## QDA itself.  Doesn't work!
qd <- qda(activity ~ ., dd)
sum(predict(qd, dd)$class != dd$activity)

## LDA plus transforms.  No better!
#for (i in 1:561) {
#    d[,i+562] <- d[,1]^2
#    dd[,i+562] <- dd[,1]^2
#}
#
#ld <- lda(activity ~ ., d)
#sum(predict(ld, dd)$class != dd$activity)

## BONUS MATERIAL

## "compression"---LDA for decreasing transmission costs, LASSO
## against LDA axes to find minimal sample set to decrease collection
## costs.
